#!/usr/bin/perl -wT
#
# リンクページ作成プログラム link.cgi ver. 1.2
#
#============================================================
# まさはる's Weblog http://homepage1.nifty.com/~masaharu/
#
# ver. 1.2 htmlで使用できない文字の変換の実装
# ver. 1.1 DB_Fileによる排他ロックの実装
# ver. 1.0 DBMハッシュを利用した最初のバージョン
#============================================================
#

use strict;
use Fcntl qw( :DEFAULT :flock );
use DB_File;
use LWP::UserAgent;
use CGI::Carp qw( fatalsToBrowser set_message );

# エラー時のデフォルトメッセージの変更
BEGIN {
	sub handle_errors {
		my $msg = shift;
		print "<h1>Software error:</h1>\n";
		print "<code>$msg</code>";
	}
	set_message( \&handle_errors );
}

#============================================================
# 設定
#============================================================

# 調査に関する設定
my $interval  = 60 * 60 * 24;	# 調査間隔[sec]
my $timeout   = 60;				# 調査時のタイムアウトまでの時間[sec]
my $agent     = 'linkbot/1.2';	# エージェント名

# 設定ファイルに関する設定
# "カテゴリー名(必須)" "サイト名(必須)" "アドレス(必須)" "コメント"\n
my $link_file = 'data/link.txt';# 調査先データのファイル名

# データベースに関する設定
my $data_dir  = 'data/';		# データベースのディレクトリ名
my $mode      = 0600;			# データベースのパーミッション

#============================================================
# 調査先データの読み込み
#============================================================

# キー:サイトのアドレス
my %new_site_ctgy;				# 値:カテゴリー名
my %new_site_name;				# 値:サイト名
my %new_site_comm;				# 値:コメント
open FILE, "<$link_file"		# 読み込み用で開く
	or die "Cannot open $link_file: $!";
foreach (<FILE>) {
	if (/^"(.+)" "(.+)" "(http:\/\/.+)" "(.*)"\n/) {
		$new_site_ctgy{$3} = &henkan($1);
		$new_site_name{$3} = &henkan($2);
		$new_site_comm{$3} = &henkan($4);
	}
}
close FILE;

#============================================================
# データベースのオープン
#============================================================

# キー:サイトのアドレス
my %SITE;						# 値:-1エラー,0未調査,1正常
my %SITE_STAT;					# 値:ステータス
my %SITE_DATE;					# 値:調査時刻
my %SITE_LAST;					# 値:更新時刻

my $db = tie %SITE, "DB_File", "${data_dir}site.db", O_CREAT | O_RDWR, $mode
	or die "Could not tie to ${data_dir}site.db: $!";
my $fd = $db->fd;
open DBM_SITE, "+<&=$fd" or die "Could not dup DBM_SITE for lock: $!";
flock DBM_SITE, LOCK_EX;		# 排他ロック
undef $db;

$db = tie %SITE_STAT, "DB_File", "${data_dir}site_stat.db", O_CREAT | O_RDWR,
	$mode or die "Could not tie to ${data_dir}site_stat.db: $!";
$fd = $db->fd;
open DBM_SITE_STAT, "+<&=$fd" or die "Could not dup DBM_SITE_STAT for lock: $!";
flock DBM_SITE_STAT, LOCK_EX;	# 排他ロック
undef $db;

$db = tie %SITE_DATE, "DB_File", "${data_dir}site_date.db", O_CREAT | O_RDWR,
	$mode or die "Could not tie to ${data_dir}site_date.db: $!";
$fd = $db->fd;
open DBM_SITE_DATE, "+<&=$fd" or die "Could not dup DBM_SITE_DATE for lock: $!";
flock DBM_SITE_DATE, LOCK_EX;	# 排他ロック
undef $db;

$db = tie %SITE_LAST, "DB_File", "${data_dir}site_last.db", O_CREAT | O_RDWR,
	$mode or die "Could not tie to ${data_dir}site_last.db: $!";
$fd = $db->fd;
open DBM_SITE_LAST, "+<&=$fd" or die "Could not dup DBM_SITE_LAST for lock: $!";
flock DBM_SITE_LAST, LOCK_EX;	# 排他ロック
undef $db;

#============================================================
# データベースへの追加と削除
#============================================================

# 追加
foreach (keys %new_site_name) {
	unless ($SITE{$_}) {
		$SITE{$_} = 0;			# 未調査のフラグ
	}
}

# 削除
foreach (keys %SITE) {
	unless (exists $new_site_name{$_}) {
		delete $SITE{$_};
		delete $SITE_STAT{$_};
		delete $SITE_DATE{$_};
		delete $SITE_LAST{$_};
	}
}

#============================================================
# リンクページの作成
#============================================================

# 正常及びエラーのアドレスのリスト作成
my @nml_adrs;					# 正常なアドレスのリスト
my @err_adrs;					# エラーのアドレスのリスト
foreach (keys %SITE) {
	if ($SITE{$_} == 1) {
		push @nml_adrs, $_;
	} elsif ($SITE{$_} == -1) {
		push @err_adrs, $_;
	}
}

# カテゴリーのリスト作成(データ順)
my @category;					# カテゴリーのリスト
if (@nml_adrs > 0) {
	foreach (@nml_adrs) {
		my $a = $_;
		my $plus = 0;
		foreach (@category) {
			if ($_ eq $new_site_ctgy{$a}) {
				$plus++;
			}
		}
		if ($plus == 0) {
			push @category, $new_site_ctgy{$a};
		}
	}
}

# カテゴリーのリストをASCII順にソート
@category = sort {
	$a cmp $b
} @category;

# 出力
print "Content-type: text/html; charset=UTF-8\n\n";
print <<END_OF_FIRST_HTML;
<!DOCTYPE html PUBLIC "-//W3C//DTD XHTML 1.0 Transitional//EN" "http://www.w3.org/TR/xhtml1/DTD/xhtml1-transitional.dtd">
<html xmlns="http://www.w3.org/1999/xhtml" xml:lang="ja" lang="ja">
<head>
<meta http-equiv="Content-Type" content="text/html; charset=UTF-8" />
<title>更新日時順リンク</title>
</head>
<body>


<!--ここは自分のサイトに合うように追記する-->


<!--ここから更新日時順のリンク-->
END_OF_FIRST_HTML

# 正常なアドレスのリストを、カテゴリー毎に更新日時順にソートして出力
if (@category > 0) {
	print "<h2 style=\"margin-bottom:20px;\">更新日時順のリンク</h2>\n";
	foreach (@category) {
		my @list;				# カテゴリー毎の正常なアドレスのリスト
		my $c = $_;
		foreach (@nml_adrs) {
			if ($new_site_ctgy{$_} eq $c) {
				push @list, $_;
			}
		}
		@list = sort {			# 更新日時の新しい順にソート
			$SITE_LAST{$b} <=> $SITE_LAST{$a} or
			$a cmp $b
		} @list;
		print "<h3>$c</h3>\n";
		print "<ul style=\"margin-bottom:40px;\">\n";
		foreach (@list) {
			if ($SITE_LAST{$_} != 0) {
				print "<li>", &localtimej($SITE_LAST{$_}), " <a href=\"$_\" target=\"_blank\">$new_site_name{$_}</a><br />$new_site_comm{$_}</li>\n";
			} else {
				print "<li><a href=\"$_\" target=\"_blank\">$new_site_name{$_}</a><br />$new_site_comm{$_}</li>\n";
			}
		}
		print "</ul>\n";
	}
}

# エラーのリンクを調査日時の新しい順にソートして出力
if (@err_adrs > 0) {
	print "<h2 style=\"margin-top:80px;\">リンク切れ</h2>\n";
	my @list = sort {			# 調査日時の新しい順にソート
		$SITE_DATE{$b} <=> $SITE_DATE{$a} or
		$a cmp $b
	} @err_adrs;
	print "<ul>\n";
	foreach (@list) {
		print "<li>調査日時: ", &localtimej($SITE_DATE{$_}), " <a href=\"$_\" target=\"_blank\">$new_site_name{$_}</a> $SITE_STAT{$_}<br />$_</li>\n";
	}
	print "</ul>\n";
}

# 出力
print <<END_OF_SECOND_HTML;
<!--ここまで-->


<!--ここは自分のサイトに合うように追記する-->


</body>
</html>
END_OF_SECOND_HTML

#============================================================
# サイトの調査
#============================================================

# LWP::UserAgentオブジェクトの生成
my $ua = LWP::UserAgent->new;
$ua->timeout($timeout);
$ua->agent($agent . ' ' . $ua->agent);

# 調査
foreach (keys %new_site_name) {

	# 以下の条件時に調査する
	# 未調査($SITE{$_} ==  0)
	# エラー($SITE{$_} == -1) && 前回の調査から設定間隔以上経過
	# 正　常($SITE{$_} ==  1) && 前回の調査から設定間隔以上経過
	if (($SITE{$_} ==  0) ||
	   (($SITE{$_} == -1) && ($SITE_DATE{$_} + $interval < time)) ||
	   (($SITE{$_} ==  1) && ($SITE_DATE{$_} + $interval < time))) {


		# 調査時刻の格納
		$SITE_DATE{$_} = time;

		# ヘッダーの取得
		my $req = HTTP::Request->new(HEAD => $_);
		my $resp = $ua->request($req);

		# 調査結果の格納 
		$SITE_STAT{$_} = $resp->status_line;
		if ($resp->is_success) {
			$SITE{$_} = 1;		# 正常のフラグ
			if ($resp->last_modified) {
				$SITE_LAST{$_} = $resp->last_modified;
			} else {
				$SITE_LAST{$_} = 0;	# 更新時刻を取得できない場合は0
			}
		} else {
			$SITE{$_} = -1;		# エラーのフラグ
			$SITE_LAST{$_} = 0;	# 更新時刻を取得できないので0
		}
	}
}
untie %SITE;
untie %SITE_STAT;
untie %SITE_DATE;
untie %SITE_LAST;

#============================================================
# 日本語のローカル時刻を返すサブルーチン
#============================================================

sub localtimej {
	my($sec,$min,$hour,$mday,$mon,$year) = localtime(shift);
	$mon  += 1;
	$year += 1900;
	$min  =~ s/^(\d)$/0$1/;
	$hour =~ s/^(\d)$/0$1/;
	$mday =~ s/^(\d)$/0$1/;
	$mon  =~ s/^(\d)$/0$1/;
	$year . "年" . $mon . "月" . $mday . "日 " . $hour . ":" . $min;
}

#============================================================
# htmlで使用できない文字の変換サブルーチン
#============================================================
# 注意: {と}の中かっこは変換時に使用するため、
#       カテゴリー名、サイト名、コメントに使用できない

sub henkan {
	$_ = shift;
	if (s/<([^<]+?)>/{$1}/g) {	# タグの外側のみ変換
		s/((?:\G|})[^{]*?)&/$1&amp;/g;
		s/((?:\G|})[^{]*?)</$1&lt;/g;
		s/((?:\G|})[^{]*?)>/$1&gt;/g;
		s/((?:\G|})[^{]*?)"/$1&quot;/g;
		s/{(.+?)}/<$1>/g;
	} else {					# タグがない場合
		s/&/&amp;/g;
		s/</&lt;/g;
		s/>/&gt;/g;
		s/"/&quot;/g;
	}
	$_;
}
